from abc import ABC, abstractmethod
from typing import Callable
from pydantic import BaseModel, Field

from zombie_nomnom.models.dice import Die
from .models.bag import DieBag


class PlayerScore(BaseModel):
    name: str
    total_brains: int = 0
    hand: list[Die] = []

    def reset(self) -> "PlayerScore":
        """
        Creates a player score that is the the same name but with no score and hand.
        """
        return PlayerScore(name=self.name)

    def clear(self) -> "PlayerScore":
        """
        Creates a player score that shares the same name and score but reset the hand.
        """
        return PlayerScore(name=self.name, total_brains=self.total_brains)

    def add_dice(self, *dice: Die) -> "PlayerScore":
        """
        Creates a playe score with the dice added to your hand.
        """
        return PlayerScore(
            name=self.name,
            hand=[*self.hand, *dice],
            total_brains=self.total_brains,
        )

    def add_brains(self, brains: int) -> "PlayerScore":
        """
        Creates a new score with a cleared hand and the brains added to your score.
        """
        return PlayerScore(
            name=self.name,
            total_brains=self.total_brains + brains,
        )


class RoundState(BaseModel):
    bag: DieBag
    player: PlayerScore
    ended: bool = False


class Command(ABC):
    """
    Used to modify round state. Cannot be used to reset game.
    """

    @abstractmethod
    def execute(self, state: RoundState) -> RoundState:  # pragma: no cover
        pass


class ZombieDieGame:
    bag: DieBag | None
    players: list[PlayerScore]
    commands: list[Command]
    bag_function: Callable[[], DieBag]
    round: RoundState | None
    current_player: int | None
    game_over: bool

    def __init__(
        self,
        players: list[PlayerScore],
        bag: DieBag | None = None,
        commands: list[Command] | None = None,
        bag_function: Callable[[], DieBag] | None = None,
    ) -> None:
        self.bag = bag
        self.commands = list(commands) if commands else []
        self.players = list(players)
        self.bag_function = bag_function or DieBag.standard_bag
        self.round = None
        self.current_player = None
        self.game_over = False

        if len(self.players) == 0:
            raise ValueError("Not enough players for the game we need at least one.")

    def reset_bag(self):
        self.bag = self.bag_function()

    def reset_players(self):  # TODO(dean): Test
        self.players = [player.reset() for player in self.players]
        self.current_player = None

    def reset_game(self):  # TODO(dean): Test
        self.reset_bag()
        self.reset_players()
        self.round = None
        self.commands = []

    def next_round(self):  # TODO(dean): Test
        self.reset_bag()
        if self.current_player is None:
            self.current_player = 0
        elif self.current_player + 1 < len(self.players):
            self.current_player = self.current_player + 1
        else:
            self.current_player = 0

    def check_for_game_over(self):
        # TODO(Milo): Check for game over.
        pass

    def process_command(self, command: Command):  # TODO(dean): Test
        if self.game_over:
            raise ValueError("Cannot command an ended game please reset game.")

        self.commands.append(command)

        if self.round is None or self.round.ended:
            self.next_round()

        self.round = command.execute(self.round)

        self.check_for_game_over()



import click
from .cli import run_game


@click.group()
def main():
    pass


@main.command("cli")
def cli():
    """
    Command to start the zombie_dice game from the command line.
    """
    run_game()

    # ask after we finish a single game assume they will quit when they want to.
    while click.confirm(text="Play another game of zombie dice?"):
        run_game()

    click.echo("Thank you for playing!!")



import random

from pydantic import BaseModel

from .dice import Die, create_die, DieColor


class DieBag(BaseModel):
    dice: list[Die]
    drawn_die: Die | None = None

    @property
    def is_empty(self):
        return len(self) == 0

    def draw_die(self) -> "DieBag":
        total = len(self)
        if total == 0:
            raise ValueError("Cannot draw_die from empty bag")

        selected_die = random.randint(0, total - 1)
        return DieBag(
            dice=[die for index, die in enumerate(self.dice) if index != selected_die],
            selected_die=self.dice[selected_die],
        )

    def __len__(self):
        return len(self.dice)

    def __bool__(self):
        return len(self) > 0

    @classmethod
    def standard_bag(cls):
        return cls(
            dice=[
                *(create_die(DieColor.GREEN) for _ in range(6)),
                *(create_die(DieColor.YELLOW) for _ in range(4)),
                *(create_die(DieColor.RED) for _ in range(3)),
            ],
        )



import pytest
from zombie_nomnom.models.dice import Die, DieColor, Face, create_die


def create_test_die() -> Die:
    # Yellow Die
    return Die(
        faces=[
            Face.BRAIN,
            Face.BRAIN,
            Face.FOOT,
            Face.FOOT,
            Face.SHOTGUN,
            Face.SHOTGUN,
        ]
    )


# Comment


def test_dice__when_we_roll_die__then_face_is_set():
    # arrange
    sut = create_test_die()
    # precheck just in case we fucked up
    assert sut.current_face is None, "Face was not set to None before we rolled it"

    # act/assert in the same line
    assert sut.roll() is not None, "Expected to get something back"
    assert sut.current_face in {
        Face.BRAIN,
        Face.FOOT,
        Face.SHOTGUN,
    }, f"Was not expected face {sut.current_face}"


def test_create_die__when_given_a_valid_color__returns_die():
    # arrange and act
    sut = create_die(DieColor.GREEN)

    # assert
    assert isinstance(sut, Die), "Expected create_die to return a Die object."


def test_create_die__when_given_invalid_color__raises_value_error():
    with pytest.raises(ValueError):
        create_die("bogus")


@pytest.mark.parametrize(
    "color, sides",
    [
        (DieColor.GREEN, [Face.BRAIN] * 3 + [Face.FOOT] * 2 + [Face.SHOTGUN] * 1),
        (DieColor.YELLOW, [Face.BRAIN] * 2 + [Face.FOOT] * 2 + [Face.SHOTGUN] * 2),
        (DieColor.RED, [Face.BRAIN] * 1 + [Face.FOOT] * 2 + [Face.SHOTGUN] * 3),
    ],
)
def test_create_die__when_given_valid_color__returns_expected_die(color, sides):
    sut = create_die(color)
    assert sut.faces == sides



import pytest
from zombie_nomnom.models.bag import DieBag
from zombie_nomnom.models.dice import Die, Face
from zombie_nomnom.engine import PlayerScore, RoundState, ZombieDieGame


@pytest.fixture
def existing_player():
    return PlayerScore(
        name="tester", total_brains=2, hand=[Die(faces=[Face.BRAIN] * 6)]
    )


@pytest.fixture
def bag_function():
    def func():
        return DieBag.standard_bag()

    return func


@pytest.fixture
def basic_game(bag_function):
    return ZombieDieGame(
        players=[PlayerScore(name="tester")],
        bag_function=bag_function,
    )


def test__player_score__reset__resets_total_brains_and_hand_but_keeps_name(
    existing_player,
):
    # act
    sut = existing_player.reset()

    # assert
    assert sut.name == existing_player.name
    assert sut.total_brains == 0
    assert sut.hand == []


def test__player_score__clear__empties_hand_but_keeps_brains_and_name(existing_player):
    # act
    sut = existing_player.clear()

    # assert
    assert sut.name == existing_player.name
    assert sut.total_brains == existing_player.total_brains
    assert sut.hand == []


def test__player_score__add_dice__joins_passed_dice_with_existing_empty_hand():
    # arrange
    existing_player = PlayerScore(name="tester", total_brains=0, hand=[])

    # act
    sut = existing_player.add_dice(Die(faces=[Face.BRAIN] * 6))

    # assert

    assert sut.hand == [Die(faces=[Face.BRAIN] * 6)]


def test__player_score__add_dice__joins_passed_dice_with_existing_filled_hand():
    # arrange
    existing_player = PlayerScore(
        name="tester", total_brains=0, hand=[Die(faces=[Face.BRAIN] * 6)] * 3
    )

    # act
    sut = existing_player.add_dice(Die(faces=[Face.BRAIN] * 6))

    # assert

    assert sut.hand == [Die(faces=[Face.BRAIN] * 6)] * 4


def test__player_score__add_brains__adds_brains_to_total_brains(existing_player):
    # act
    sut = existing_player.add_brains(5)

    # assert
    assert sut.total_brains == 7


def test__zombie_die_game__init_raises_value_error_when_players_is_zero():
    with pytest.raises(ValueError):
        ZombieDieGame(
            players=[],
        )


def test__zombie_die_game__reset_bag_calls_passed_bag_function_and_sets_to_standard_bag(
    basic_game,
):
    # act
    assert basic_game.bag is None  # make sure there is no bag before we reset
    basic_game.reset_bag()

    # assert
    assert isinstance(basic_game.bag, DieBag)
    assert basic_game.bag.dice == DieBag.standard_bag().dice



"""
The cli version of zombie_dice this is where we manage the state of the game and how we 
format commands from the cli to apply to the engine and render that to the user.
"""

import click

from .engine import PlayerScore, ZombieDieGame


def run_game():
    game = setup_game()
    # TODO(Milo): Figure out how to do the thingy


def setup_game() -> ZombieDieGame:
    names = prompt_list(
        "Enter Player Name",
        _type=str,
        confirmation_prompt="Add Another Player?",
    )
    # TODO(Milo): Figure out a bunch of game types to play that we can use as templates for the die.
    return ZombieDieGame(
        players=[PlayerScore(name=name) for name in names],
    )


def prompt_list(
    prompt: str,
    _type: type,
    confirmation_prompt: str = "Add Another?",
) -> list:
    inputs = []
    inputs.append(click.prompt(prompt, type=_type))

    while click.confirm(confirmation_prompt):
        inputs.append(click.prompt(prompt, type=_type))
    return inputs



from enum import Enum
from pydantic import BaseModel, Field
import random


class Face(str, Enum):
    BRAIN = "BRAIN"
    FOOT = "FOOT"
    SHOTGUN = "SHOTGUN"


class DieColor(str, Enum):
    RED = "RED"
    YELLOW = "YELLOW"
    GREEN = "GREEN"


class Die(BaseModel):
    faces: list[Face] = Field(min_length=6, max_length=6)
    current_face: Face | None = None

    def roll(self) -> Face:
        self.current_face = random.choice(self.faces)
        return self.current_face


_dice_face_mapping = {
    DieColor.RED: {Face.BRAIN: 1, Face.FOOT: 2, Face.SHOTGUN: 3},
    DieColor.YELLOW: {Face.BRAIN: 2, Face.FOOT: 2, Face.SHOTGUN: 2},
    DieColor.GREEN: {Face.BRAIN: 3, Face.FOOT: 2, Face.SHOTGUN: 1},
}


def create_die(color: DieColor) -> Die:
    if color not in _dice_face_mapping:
        raise ValueError(f"Unknown Die Color: {color}")

    mapped_color = _dice_face_mapping[color]
    faces = []
    for face, amount in mapped_color.items():
        faces.extend([face] * amount)
    return Die(faces=faces)



import pytest
from zombie_nomnom.models.bag import DieBag


def test_die_bag__when_creating_standard_bag__contains_13_dice():
    bag = DieBag.standard_bag()
    assert len(bag) == 13


def test_die_bag__when_using_bool_check_and_contains_dice__returns_true():
    bag = DieBag.standard_bag()
    assert bag, "Not truthy when has dice"


def test_die_bag__when_using_bool_check_and_does_not_contain_dice__returns_false():
    bag = DieBag(dice=[])
    assert not bag, "Should not be true"


def test_die_bag__when_using_bool_check_for_bag__should_be_opposite_of_is_empty():
    bag = DieBag(dice=[])
    assert bool(bag) == (not bag.is_empty), "Should not match is empty"


def test_die_bag__when_drawing_die_and_bag_empty__raises_value_error():
    bag = DieBag(dice=[])

    with pytest.raises(ValueError):
        bag.draw_die()
